import inspect
from .exceptions import ValidationError

_validators = {}
_logger = None


class Validator:
    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Validator.set_logger()
        cls.retrieve_fields()
        _validators[cls.__name__] = cls

    @staticmethod
    def set_logger():
        global _logger
        if _logger is None:
            from .app import _app

            _logger = _app.logger

    @classmethod
    def retrieve_fields(cls):
        dct = cls.__dict__
        validators = {}
        [
            validators.update({field_name: field_obj})
            for field_name, field_obj in dct.items()
            if not field_name.startswith("__")
        ]
        cls.validators = validators
        [delattr(cls, field_name) for field_name in validators]

    @classmethod
    def validated_message(cls, message):
        if not type(message) in [dict, list]:
            error = (
                f"Unexpected message. Accepted dict or list but got {type(message)}."
            )
            _logger.error(error)
            raise ValidationError(error)
        if type(message) is list:
            if not cls.__many:
                error = (
                    f"Unexpected message, expected dict  got {type(message)}."
                    "You can set many=True in validator if you need to handle list of dicts"
                )
                _logger.error(error)
                raise ValidationError(error)
            result = []
            for m in message:
                result.append(cls._validate_message(cls, m))
            return result
        message = cls._validate_message(message)
        # print(f'message {message} is OK {cls.validators}')
        return message

    @classmethod
    def _validate_message(cls, message):
        if not type(message) is dict:
            error = "Unexpected message, expected dict"
            _logger.error(error)
            raise ValidationError(error)
        for key, field_obj in cls.validators.items():
            if issubclass(field_obj.type.__base__, Validator):
                field_obj.type.validated_message(message[key])
                continue
            if key not in message and not hasattr(field_obj, "default"):
                error = f'Expected field "{key}" not found'
                _logger.error(error)
                raise ValidationError(error)
            elif key not in message and hasattr(field_obj, "default"):
                message[key] = field_obj.default
            if message[key] is None and field_obj.null is False:
                error = f'Wrong value None for field "{key}" (because null=False)'
                _logger.error(error)
                raise ValidationError(error)
            if message[key] is not None and not isinstance(
                message[key], field_obj.type
            ):
                error = f'Expected {field_obj.type} type of field "{key}" but got {type(message[key])} instead'
                _logger.error(error)
                raise ValidationError(error)
            continue
        return message


class Field:
    def __init__(self, many=False, null=False, **kwargs):
        # todo: check all incoming attrs
        self.__many = many
        kwargs["null"] = null
        self.validate_field(kwargs)
        self.__dict__.update(kwargs)

    @staticmethod
    def validate_field(kwargs):
        if "type" not in kwargs:
            error = "type required in Field"
            _logger.error(error)
            raise ValidationError(error)
        if (
            "default" in kwargs
            and kwargs["default"] is None
            and ("null" not in kwargs or kwargs["null"] is False)
        ):
            error = "You have to set null=True first if you want to set default=None"
            _logger.error(error)
            raise ValidationError(error)
        if (
            "default" in kwargs
            and kwargs["default"] is not None
            and not isinstance(kwargs["default"], kwargs["type"])
        ):
            error = f'Your default type is {type(kwargs["default"])} but expected {kwargs["type"]}'
            _logger.error(error)
            raise ValidationError(error)
        if not kwargs["type"] in [str, int, float, list, dict, bool]:
            all_classes = inspect.getmro(kwargs["type"])
            if len(all_classes) > 1:
                validator_name = all_classes[0].__name__
                if validator_name not in _validators:
                    error = f"Validator {validator_name} hasn't registered yet. You have to register in first"
                    _logger.error(error)
                    raise ValidationError(error)
                parent_name = all_classes[1].__name__
                if not parent_name == "Validator":
                    error = (
                        'You have to inherit your validator from the class "Validator"'
                    )
                    _logger.error(error)
                    raise ValidationError(error)
            else:
                error = (
                    f'Invalid data type {kwargs["type"]} for field.'
                    f" Only JSON data types or another Validator class allowed"
                )
                _logger.error(error)
                raise ValidationError(error)



import asyncio
import os
import shutil
import json
import random
import time
import typing
from urllib.parse import urljoin

import requests
import websocket
from nats.aio.client import Client as NATS

from .exceptions import TestClientError
from .utils.helper import start_process
from panini.managers.nats_client import Msg


# Annotations for `Session.request()`
Cookies = typing.Union[
    typing.MutableMapping[str, str], requests.cookies.RequestsCookieJar
]
Params = typing.Union[bytes, typing.MutableMapping[str, str]]
DataType = typing.Union[bytes, typing.MutableMapping[str, str], typing.IO]
TimeOut = typing.Union[float, typing.Tuple[float, float]]
FileType = typing.MutableMapping[str, typing.IO]
AuthType = typing.Union[
    typing.Tuple[str, str],
    requests.auth.AuthBase,
    typing.Callable[[requests.Request], requests.Request],
]


class HTTPSessionTestClient(requests.Session):
    __test__ = False

    def __init__(self, base_url: str = "http://127.0.0.1:8080"):
        super(HTTPSessionTestClient, self).__init__()
        self.base_url = base_url

    def request(  # type: ignore
        self,
        method: str,
        url: str,
        params: Params = None,
        data: DataType = None,
        headers: typing.MutableMapping[str, str] = None,
        cookies: Cookies = None,
        files: FileType = None,
        auth: AuthType = None,
        timeout: TimeOut = None,
        allow_redirects: bool = None,
        proxies: typing.MutableMapping[str, str] = None,
        hooks: typing.Any = None,
        stream: bool = None,
        verify: typing.Union[bool, str] = None,
        cert: typing.Union[str, typing.Tuple[str, str]] = None,
        json: typing.Any = None,
    ) -> requests.Response:
        url = urljoin(self.base_url, url)
        return super().request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )


def get_logger_files_path(folder: str = "test_logs", remove_if_exist: bool = False):
    testing_directory_path = os.getcwd()
    testing_logs_directory_path = os.path.join(testing_directory_path, folder)
    if remove_if_exist:
        if os.path.exists(testing_logs_directory_path):
            shutil.rmtree(testing_logs_directory_path)

    return testing_logs_directory_path


def is_subject_matches_pattern(subject: str, pattern: str) -> bool:
    subject_parts = subject.split(".")
    pattern_parts = pattern.split(".")
    if ">" in pattern:
        assert pattern[-1] == ">", "> must can be only the last element of subject!"
        gt_position = pattern_parts.index(">")
        return is_subject_matches_pattern(
            ".".join(subject_parts[:gt_position]),
            ".".join(pattern_parts[:gt_position]),
        )

    if ">" in subject_parts:
        return False

    if len(subject_parts) != len(pattern_parts):
        return False

    for index, part in enumerate(pattern_parts):
        if part == "*":
            pass
        elif part != subject_parts[index]:
            return False

    return True


class AsyncTestClient:
    __test__ = False

    def __init__(
        self,
        run_panini: typing.Callable = None,
        run_panini_args: list = None,
        run_panini_kwargs: dict = None,
        run_panini_timeout: float = 5,
        panini_service_name: str = "*",
        panini_client_id: str = "*",
        logger_files_path: str = "test_logs",
        use_web_server: bool = False,
        use_web_socket: bool = False,
        base_web_server_url: str = "http://127.0.0.1:8080",
        nats_host: str = "127.0.0.1",
        nats_port: int = 4222,
        name: str = "__".join(
            [
                "test_client",
                str(random.randint(1, 10000000)),
                str(random.randint(1, 10000000)),
            ]
        ),
        listen_subjects_callbacks: dict = None,
    ):
        self.run_panini = run_panini
        self.run_panini_args = run_panini_args or []
        self.run_panini_kwargs = run_panini_kwargs or {}
        self.run_panini_timeout = run_panini_timeout
        self.listen_subjects_callbacks = listen_subjects_callbacks or {}
        self.panini_service_name = panini_service_name
        self.panini_client_id = panini_client_id
        self.logger_files_path = logger_files_path
        self.base_web_server_url = base_web_server_url
        self.use_web_server = use_web_server
        self.use_web_socket = use_web_socket
        self.name = name
        self.nats_client = NATS()
        self.nats_client.msg_class = Msg
        self.nats_host = nats_host
        self.nats_port = nats_port
        self._listen_subjects_count_calls = {}

        if use_web_server:
            self.http_session = HTTPSessionTestClient(base_url=base_web_server_url)

        if use_web_socket:
            self.websocket_session = websocket.WebSocket()

        self.panini_process = None

    @staticmethod
    def _dict_to_bytes(message: dict) -> bytes:
        return json.dumps(message).encode("utf-8")

    @staticmethod
    def _bytes_to_dict(payload: bytes) -> dict:
        return json.loads(payload)

    @staticmethod
    def wrap_run_panini(
        run_panini,
        run_panini_args: list,
        run_panini_kwargs: dict,
        logger_files_path: str,
        use_error_middleware: bool,
    ):
        from .utils.logger import get_logger

        test_logger = get_logger("panini")
        # set the panini testing data in os.environ
        os.environ["PANINI_TEST_MODE"] = "true"
        os.environ["PANINI_TEST_MODE_USE_ERROR_MIDDLEWARE"] = (
            "true" if use_error_middleware else "false"
        )
        testing_logger_files_path = (
            get_logger_files_path(logger_files_path)
            if not os.path.isabs(logger_files_path)
            else logger_files_path
        )

        os.environ["PANINI_TEST_LOGGER_FILES_PATH"] = testing_logger_files_path

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            run_panini(*run_panini_args, **run_panini_kwargs)
        except Exception as e:
            test_logger.exception(f"Run panini error: {e}")

    async def start(self, is_daemon: bool = True, use_error_middleware=True):
        await self.nats_client.connect(
            f"{self.nats_host}:{self.nats_port}", name=self.name
        )
        for subject, callback in self.listen_subjects_callbacks.items():
            await self.nats_client.subscribe(subject, cb=callback)

        if self.run_panini is not None:

            panini_started_future = asyncio.Future()

            def panini_started(msg):
                panini_started_future.set_result(msg)

            sid = await self.nats_client.subscribe(
                f"panini_events.{self.panini_service_name}.{self.panini_client_id}.started",
                cb=panini_started,
            )
            await self.nats_client.auto_unsubscribe(sid, 1)

            self.panini_process = start_process(
                self.wrap_run_panini,
                args=(
                    self.run_panini,
                    self.run_panini_args,
                    self.run_panini_kwargs,
                    self.logger_files_path,
                    use_error_middleware,
                ),
                daemon=is_daemon,
            )

            try:
                await asyncio.wait_for(panini_started_future, self.run_panini_timeout)
            except Exception:
                raise TestClientError(
                    "TestClient was waiting panini to start, but panini does not started"
                )

            if self.use_web_server:
                pass  # TODO: understand, why don't we need to wait for web_server

        return self

    async def stop(self):
        await self.nats_client.drain()
        await self.nats_client.close()

        if self.panini_process is not None:
            self.panini_process.kill()

        if hasattr(self, "http_session"):
            self.http_session.close()

        if hasattr(self, "websocket_session"):
            self.websocket_session.close()

    async def publish(self, subject: str, message: dict, reply_to: str = "") -> None:
        message = self._dict_to_bytes(message)
        if reply_to is not None:
            await self.nats_client.publish_request(subject, reply_to, message)
        else:
            await self.nats_client.publish(subject, message)

    async def request(self, subject: str, message: dict, timeout: int = 1) -> dict:
        response = await self.nats_client.request(
            subject=subject, payload=self._dict_to_bytes(message), timeout=timeout
        )
        return self._bytes_to_dict(response.data)

    async def subscribe(
        self,
        subject: str,
        callback: typing.Callable,
        queue: str = "",
        max_messages: typing.Optional[int] = 0,
    ):
        return await self.nats_client.subscribe(
            subject=subject,
            cb=callback,
            queue=queue,
            max_msgs=max_messages,
        )

    def total_count_subject_calls(self, count_calls: dict = None):
        if count_calls is None:
            count_calls = self._listen_subjects_count_calls
        return sum(count_calls.values())

    def count_subject_calls(self, subject: str, count_calls: dict = None):
        """Count calls for specific subject (this subject is used as pattern for subject - supports * and >)"""
        if count_calls is None:
            count_calls = self._listen_subjects_count_calls
        res_count = 0
        for listen_subject, count in count_calls.items():
            if is_subject_matches_pattern(listen_subject, subject):
                res_count += count

        return res_count

    def clear_listen_subjects_count_calls(self):
        """Clean up all counts for subject called in AsyncTestClient"""
        self._listen_subjects_count_calls = {}

    async def wait(
        self,
        count: int = 0,
        timeout: float = 1,
        subject: str = None,
        subjects: dict = None,
    ) -> None:
        """
        Waits for test client subjects to be triggered & handled.
        count: how much calls wait
        timeout: maximum seconds to wait
        subject: which subject to wait
        subjects: dict of subjects, with subject as key and count_to_wait as value
        """
        start_time = time.time()

        while time.time() - start_time < timeout:
            if subject is not None:
                if self.count_subject_calls(subject) >= count:
                    return
            elif subjects is not None:
                if all(
                    [
                        self.count_subject_calls(subj) >= count_call
                        for subj, count_call in subjects.items()
                    ]
                ):
                    return
            elif self.total_count_subject_calls() >= count:
                return
            await asyncio.sleep(0)

        # wait was not successful
        if subject is not None:
            report_msg = f"subject: {subject} was called {self.count_subject_calls(subject)} times"
        elif subjects is not None:
            subjects_calls = {subj: self.count_subject_calls(subj) for subj in subjects}
            report_msg = f"subjects were called {subjects_calls} times"
        else:
            report_msg = f"total_count_subject_calls={self.total_count_subject_calls()}"
        raise asyncio.TimeoutError(
            f"Timeout while waiting for listen_subjects to be called! Params - count: {count}, timeout: {timeout}, subject: {subject}, subjects: {subjects}. Actual state: {report_msg}"
        )

    def listen(self, subject: str):
        def decorator(func):
            assert isinstance(subject, str), "Subject must be only in str format"

            async def wrapper(incoming_message):
                assert isinstance(
                    incoming_message, Msg
                ), "Incoming message must be instance of Msg!"
                incoming_message_data = self._bytes_to_dict(incoming_message.data)

                msg = Msg(
                    subject=incoming_message.subject,
                    data=incoming_message_data,
                    reply=incoming_message.reply,
                    sid=incoming_message.sid,
                )
                if asyncio.iscoroutinefunction(func):
                    wrapper_response = await func(msg)
                else:
                    wrapper_response = func(msg)
                if wrapper_response is not None and incoming_message.reply != "":
                    await self.nats_client.publish(
                        subject=incoming_message.reply,
                        payload=self._dict_to_bytes(wrapper_response),
                    )

                self._listen_subjects_count_calls[msg.subject] = (
                    self._listen_subjects_count_calls.get(msg.subject, 0) + 1
                )

            self.listen_subjects_callbacks[subject] = wrapper

            return wrapper

        return decorator



import asyncio

from panini import exceptions
from panini.exceptions import NotReadyError, ValidationError


class EventManager:
    """
    Collect all functions from each module wrapped by @app.subscription or @EventManager.subscribe
    """

    def __init__(self):
        self._subscriptions = {}

    @property
    def subscriptions(self):
        return self._subscriptions

    def listen(
        self,
        subject: list or str,
        validator: type = None,
        data_type="json.loads",
    ):
        def wrapper(function):
            function = self.wrap_function_by_validator(function, validator)
            if type(subject) is list:
                for t in subject:
                    self._check_subscription(t)
                    self._subscriptions[t].append(function)
            else:
                self._check_subscription(subject)
                self._subscriptions[subject].append(function)
            function.data_type = data_type
            return function
        return wrapper

    def wrap_function_by_validator(self, function, validator):
        def validate_message(msg):
            try:
                if validator is not None:
                    validator.validated_message(msg.data)
            except exceptions.ValidationError as se:
                error = f"subject: {msg.subject} error: {str(se)}"
                return {"success": False, "error": error}
            except Exception as e:
                raise ValidationError(e)
            return msg

        def wrapper(msg):
            validate_message(msg)
            return function(msg)

        async def wrapper_async(msg):
            validate_message(msg)
            return await function(msg)

        if asyncio.iscoroutinefunction(function):
            return wrapper_async
        else:
            return wrapper

    def _check_subscription(self, subscription):
        if subscription not in self._subscriptions:
            self._subscriptions[subscription] = []



import time

from prometheus_client import CollectorRegistry, Histogram, Counter, push_to_gateway
from prometheus_client.utils import INF
from nats.aio.errors import ErrTimeout

from . import Middleware
from ..app import get_app


DEFAULT_BUCKETS = (
    0.001,
    0.005,
    0.01,
    0.025,
    0.05,
    0.075,
    0.1,
    0.25,
    0.5,
    0.75,
    1.0,
    2.5,
    5.0,
    7.5,
    10.0,
    INF,
)

DEFAULT_LABELS = (
    "app_name",
    "client_id",
    "subject",
    "status",
)


class PrometheusMonitoringMiddleware(Middleware):
    def __init__(
        self,
        pushgateway_url: str = "localhost:9091",
        app=None,
        frequency: float = 10.0,
        job: str = "microservices_activity",
        buckets: tuple = DEFAULT_BUCKETS,
        labels: tuple = DEFAULT_LABELS,
    ):
        if app is None:
            app = get_app()
        assert app is not None, "App must be initialized before this middleware"
        self.pushgateway_url = pushgateway_url
        self.registry = CollectorRegistry()
        self.app_name = app.service_name
        self.app_client_id = app.client_id
        self.buckets = buckets
        assert set(labels).issubset(
            set(DEFAULT_LABELS)
        ), f"Labels must be in {DEFAULT_LABELS}"
        self.labels = labels
        self.listen_counter = self.create_listen_counter()
        self.listen_latency_histogram = self.create_listen_latency_histogram()

        @app.timer_task(frequency)
        async def push_to_prometheus():
            push_to_gateway(pushgateway_url, job=job, registry=self.registry)

    def create_listen_counter(self):
        counter = Counter(
            "panini_listen_total",
            "Listen duration in seconds",  # description
            registry=self.registry,
            labelnames=self.labels,
        )
        return counter

    def create_listen_latency_histogram(self):
        histogram = Histogram(
            "panini_listen_latency_seconds",
            "Listen duration in seconds",  # description
            registry=self.registry,
            buckets=self.buckets,
            labelnames=self.labels,
        )
        return histogram

    def monitor_listen(self, start_time: float, labels: dict):
        duration = time.time() - start_time
        labels = {
            label_key: label_value
            for label_key, label_value in labels.items()
            if label_key in self.labels
        }
        self.listen_counter.labels(**labels).inc()
        self.listen_latency_histogram.labels(**labels).observe(duration)

    async def listen_any(self, msg, callback):
        labels = {
            "app_name": self.app_name,
            "client_id": self.app_client_id,
            "subject": msg.subject,
        }

        start_time = time.time()

        if "status" in self.labels:
            try:
                response = await callback(msg)
            except Exception:
                labels["status"] = "failure"
                self.monitor_listen(start_time, labels)
                raise
            else:
                labels["status"] = "success"
        else:
            response = await callback(msg)

        self.monitor_listen(start_time, labels)

        return response



import asyncio
import logging
import os
import typing
import uuid
from types import CoroutineType

from aiohttp import web

from panini.managers.nats_client import NATSClient
from .exceptions import InitializingEventManagerError

from .http_server.http_server_app import HTTPServer
from .managers.event_manager import EventManager
from .managers.task_manager import TaskManager
from .middleware.error import ErrorMiddleware
from .utils import logger
from .utils.helper import (
    get_app_root_path,
    create_client_code_by_hostname,
)

_app = None


class App:
    def __init__(
            self,
            host: str = None,
            port: int or str = None,
            service_name: str = "panini_microservice_" + str(uuid.uuid4())[:10],
            servers: list = None,
            client_id: str = None,
            reconnect: bool = False,
            max_reconnect_attempts: int = 60,
            reconnecting_time_sleep: int = 2,
            allocation_queue_group: str = "",

            logger_required: bool = True,
            logger_files_path: str = None,
            logger_in_separate_process: bool = False,
            pending_bytes_limit=65536 * 1024 * 10,
            **kwargs
    ):
        """
        :param host: NATS broker host
        :param port: NATS broker port
        :param service_name: Name of microservice
        :param servers: Alternative to NATS broker host+NATS broker port. Allowed establish connection to multiple NATS brokers
        :param client_id: id of microservice, name and client_id used for NATS client name generating
        :param tasks: List of additional tasks
        :param reconnect: allows reconnect if connection to NATS has been lost
        :param max_reconnect_attempts: number of reconnect attempts
        :param reconnecting_time_sleep: pause between reconnection
        :param subscribe_subjects_and_callbacks: if you need to subscribe additional
                                        subjects(except subjects from event.py).
                                        This way doesn't support validators
        :param allocation_queue_group: name of NATS queue for distributing incoming messages among many NATS clients
                                    more detailed here: https://docs.nats.io/nats-concepts/queue
        :param listen_subject_only_if_include:   if not None, client will subscribe
                                                only to subjects that include these key words
        :param listen_subject_only_if_exclude:   if not None, client will not subscribe
                                                 to subjects that include these key words
        :param web_app: web.Application:       custom aiohttp app that you can create separately from panini.
                            if you set this argument client will only run this aiohttp app without handling
        :param web_host: Web application host
        :param web_port: Web application port
        :param logger_required: Is logger required for the project (if not - EmptyLogger will be provided)
        :param logger_files_path: main path for logs
        :param logger_in_separate_process: use log in the same or in different process
        """

        try:
            # client_id initialization
            if client_id is None:
                self.client_id = create_client_code_by_hostname(service_name)
            else:
                self.client_id = client_id
            os.environ["CLIENT_ID"] = self.client_id

            self.service_name = service_name

            self._event_manager = EventManager()
            self._task_manager = TaskManager()
            self.nats = NATSClient(
                host=host,
                port=port,
                servers=servers,
                client_id=self.client_id,
                allow_reconnect=reconnect,
                queue=allocation_queue_group,
                max_reconnect_attempts=max_reconnect_attempts,
                reconnecting_time_wait=reconnecting_time_sleep,
                pending_bytes_limit=pending_bytes_limit,
                **kwargs
            )

            self.app_root_path = get_app_root_path()
            self.logger_required = logger_required
            self.logger_in_separate_process = logger_in_separate_process

            # check, if TestClient is running
            if os.environ.get("PANINI_TEST_MODE"):
                self.logger_files_path = os.environ.get(
                    "PANINI_TEST_LOGGER_FILES_PATH", "test_logs"
                )
            else:
                self.logger_files_path = (
                    logger_files_path if logger_files_path else "logs"
                )
            self.logger = logger.Logger(None)

            if self.logger_required:
                self.set_logger(
                    self.service_name,
                    self.app_root_path,
                    self.logger_files_path,
                    False,
                    self.client_id,
                )

            self.logger_process = None
            self.log_stop_event = None
            self.log_listener_queue = None
            self.change_logger_config_listener_queue = None

            self.http_server = None
            self.http = None

            self.task = self._task_manager.register_single_task
            self.timer_task = self._task_manager.register_interval_task

            global _app
            _app = self

        except InitializingEventManagerError as e:
            error = f"App.event_registrar critical error: {str(e)}"
            raise InitializingEventManagerError(error)

    def setup_web_server(self, host=None, port=None, web_app=None, params: dict = None):
        """
        Setup server and run with NATS when called app.start()
        """
        self.http = web.RouteTableDef()  # for http decorator
        if web_app:
            self.http_server = HTTPServer(routes=self.http, web_app=web_app, web_server_params=params)
        else:
            self.http_server = HTTPServer(routes=self.http, host=host, port=port, web_server_params=params)

    def add_filters(self, include: list = None, exclude: list = None):
        """
        Allows to listen subject only if subject include or exclude string from lists in agrs
        """
        return self.nats.add_filters(include, exclude)

    def set_logger(
            self,
            service_name,
            app_root_path,
            logger_files_path,
            in_separate_process,
            client_id
    ):
        if in_separate_process:
            (
                self.log_listener_queue,
                self.log_stop_event,
                self.logger_process
            ) = logger.set_logger(
                service_name,
                app_root_path,
                logger_files_path,
                in_separate_process,
                client_id,
            )
        else:
            logger.set_logger(
                service_name,
                app_root_path,
                logger_files_path,
                in_separate_process,
                client_id,
            )
        self.logger.logger = logging.getLogger(service_name)

    def listen(
            self,
            subject: list or str,
            validator: type = None,
            data_type="json.loads"
    ):
        return self._event_manager.listen(subject, validator, data_type)

    async def publish(
            self,
            subject: str,
            message,
            reply_to: str = None,
            force: bool = False,
            data_type: type or str = "json.dumps"
    ):
        return await self.nats.publish(subject, message, reply_to, force, data_type)

    def publish_sync(
            self,
            subject: str,
            message,
            reply_to: str = None,
            force: bool = False,
            data_type: type or str = "json.dumps",
    ):
        return self.nats.publish_sync(subject, message, reply_to, force, data_type)

    async def request(
            self,
            subject: str,
            message,
            timeout: int = 10,
            data_type: type or str = "json.dumps",
            callback: typing.Callable = None
    ):
        return await self.nats.request(subject, message, timeout, data_type, callback)

    def request_sync(
            self,
            subject: str,
            message,
            timeout: int = 10,
            data_type: type or str = "json.dumps",
            callback: typing.Callable = None,
    ):
        return self.nats.request_sync(subject, message, timeout, data_type, callback)

    def subscribe_new_subject_sync(self, subject: str, callback: CoroutineType, **kwargs):
        return self.nats.subscribe_new_subject_sync(subject, callback, **kwargs)

    async def subscribe_new_subject(
            self,
            subject: str,
            callback: CoroutineType,
            init_subscription=False,
            is_async=False,
            data_type=None
    ):
        return await self.nats.subscribe_new_subject(subject, callback, init_subscription, is_async, data_type)

    def unsubscribe_subject_sync(self, subject: str):
        return self.nats.unsubscribe_subject_sync(subject)

    async def unsubscribe_subject(self, subject: str):
        return await self.nats.unsubscribe_subject(subject)

    def unsubscribe_ssid_sync(self, ssid: int, subject: str = None):
        return self.nats.unsubscribe_ssid_sync(ssid, subject)

    async def unsubscribe_ssid(self, ssid: int, subject: str = None):
        return await self.nats.unsubscribe_ssid(ssid, subject)

    def add_middleware(self, cls, *args, **kwargs):
        return self.nats.middleware_manager.add_middleware(cls, *args, **kwargs)

    def _start_event(self):
        asyncio.ensure_future(
            self.nats.client.publish(
                f"panini_events.{self.service_name}.{self.client_id}.started",
                b"{}",
            )
        )

    def start(self):
        if (
                os.environ.get("PANINI_TEST_MODE")
                and os.environ.get("PANINI_TEST_MODE_USE_ERROR_MIDDLEWARE", "false")
                == "true"
        ):
            def exception_handler(e, **kwargs):
                self.logger.exception(f"Error: {e}, for kwargs: {kwargs}")
                raise

            self.add_middleware(
                ErrorMiddleware, error=Exception, callback=exception_handler
            )
        if self.logger_required and self.logger_in_separate_process:
            self.set_logger(
                self.service_name,
                self.app_root_path,
                self.logger_files_path,
                True,
                self.client_id,
            )

        self.nats.set_listen_subjects_callbacks(self._event_manager.subscriptions)
        self.nats.start()

        loop = asyncio.get_event_loop()

        self._start_event()

        if self.http_server:
            self.http_server.start_server()

        tasks = asyncio.all_tasks(loop)
        self._task_manager.create_tasks()
        loop.run_until_complete(asyncio.gather(*tasks))


def get_app() -> App:
    return _app



import asyncio
import time
from typing import Callable

from panini.exceptions import InitializingTaskError


class TaskManager:
    """
    Collect all functions from each module wrapped by @app.task or @TaskManager.task
    """

    def __init__(self):
        self._tasks = []
        self._loop = asyncio.get_event_loop()

    @property
    def tasks(self):
        return self._tasks

    def __call__(self, interval: float or int = None, **kwargs):
        def wrapper(task):
            if interval:
                self.register_interval_task(interval, task)
            else:
                self.register_single_task(task)
            self.register_single_task(task)
            return task

        return wrapper

    def register_single_task(self):
        def wrapper(task):
            self._check_task(task)
            self._tasks.append(task)
        # self._loop.create_task(task())
        return wrapper

    def register_interval_task(self, interval: int or float):
        def wrapper(task):
            self._check_task(task)
            interval_task = self.wrapper_for_interval_task(interval, task)
            self._tasks.append(interval_task)
            # self._loop.create_task(interval_task())
        return wrapper

    def wrapper_for_interval_task(self, interval, task):
        async def wrapper(**kwargs):
            while True:
                try:
                    start = time.time()
                    await task(**kwargs)
                    duration = time.time() - start

                    if duration < interval:
                        await asyncio.sleep(interval - duration)

                except InitializingTaskError:
                    # TODO: warning log
                    pass

        return wrapper

    def _check_task(self, task):
        if not asyncio.iscoroutinefunction(task):
            raise InitializingTaskError("Only coroutine tasks allowed")

    def create_tasks(self):
        loop = asyncio.get_event_loop()
        for task in self._tasks:
            loop.create_task(task())



import asyncio
import json
import threading
import time

from nats.aio.client import Client as NATS


def _dict_to_bytes(message: dict) -> bytes:
    return json.dumps(message).encode("utf-8")


def _bytes_to_dict(payload: bytes) -> dict:
    return json.loads(payload)


class EmulatorClient(threading.Thread):
    def __init__(
        self,
        filepath: str,
        prefix: str,
        app_name: str,
        emulate_timeout: bool = True,
        compare_output: bool = False,
        max_timeout_after_start: float = 20.0,
    ):
        threading.Thread.__init__(self)

        self._name = "emulator_client" + prefix
        self._filepath = filepath

        self._publish_queue = []
        self._listen_queues = {}

        self._emulate_timeout = emulate_timeout
        self._compare_output = compare_output
        self._max_timeout_after_start = max_timeout_after_start
        self._client = NATS()

        self._app_name = app_name

        self._prefix = prefix
        self._reply_to_suffix = "reply"
        self._subscriptions = []

        self._is_app_started = False
        self._is_emulator_ready = False

        self._load()

    def _load(self):
        with open(self._filepath) as file:
            lines = file.readlines()

        for line in lines:
            event = json.loads(line)
            event_type = event["event_type"]
            subject = self._prefix + "." + event["subject"]

            if event_type.startswith("listen"):
                self._publish_queue.append(event)

            elif event_type.startswith("send"):
                if subject not in self._listen_queues:
                    self._listen_queues[subject] = []

                self._listen_queues[subject].append(event)

    async def _mock_request(self, message):
        subject = message.subject
        reply_to = message.reply
        body = _bytes_to_dict(message.data)
        event = self._listen_queues[subject].pop(0)
        if self._compare_output:
            assert body == event["message"], f"'{body}' vs '{event['message']}'"

        if event["event_type"].endswith("request"):
            await self._client.publish(reply_to, _dict_to_bytes(event["response"]))

    def listen(self, subject: str):
        def decorator(func):
            assert isinstance(subject, str)

            def wrapper(incoming_response):
                wrapper_response = func(
                    topic=incoming_response.subject,
                    message=_bytes_to_dict(incoming_response.data),
                )
                if wrapper_response is not None and incoming_response.reply != "":
                    self._client.publish(incoming_response.reply, wrapper_response)

            self._subscriptions.append([self._prefix + "." + subject, wrapper])

            return wrapper

        return decorator

    def run(self):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._run())

    async def _on_app_started(self, message):
        self._is_app_started = True

    async def _run(self):
        await self._client.connect()

        for subject in self._listen_queues:
            await self._client.subscribe(subject, cb=self._mock_request)

        for subject, callback in self._subscriptions:
            await self._client.subscribe(subject, cb=callback)

        subject = f"{self._prefix}.panini_events.{self._app_name}.*.started"
        await self._client.subscribe(subject, cb=self._on_app_started)

        self._is_emulator_ready = True
        await self._wait_for_app_to_start()
        # publish all the events
        await self._run_publish()

        await self._wait_after()

    async def _run_publish(self):
        last_timestamp = None
        while len(self._publish_queue) > 0:
            event = self._publish_queue.pop(0)

            event_type = event["event_type"]
            subject = self._prefix + "." + event["subject"]
            timestamp = event["timestamp"]
            message = _dict_to_bytes(event["message"])

            if event_type.endswith("publish"):
                await self._client.publish(subject, message)

            elif event_type.endswith("request"):
                response = await self._client.request(subject, message, timeout=5)

                if self._reply_to_suffix:
                    await self._client.publish(
                        f"{subject}.{self._reply_to_suffix}",
                        _dict_to_bytes(event["response"]),
                    )

                if self._compare_output:
                    assert json.loads(response.data) == event["response"]

            if self._emulate_timeout:
                if last_timestamp:
                    wait = timestamp - last_timestamp
                    time.sleep(wait)

                last_timestamp = timestamp

    def wait_for_readiness(self):
        while not self._is_emulator_ready:
            time.sleep(0.1)

    async def _wait_for_app_to_start(self):
        while not self._is_app_started:
            await asyncio.sleep(0.1)

    async def _wait_after(self):
        start = time.time()
        while True:
            if sum(len(queue) for queue in self._listen_queues.values()) == 0:
                break

            if (
                self._max_timeout_after_start
                and time.time() - start > self._max_timeout_after_start
            ):
                break

            await asyncio.sleep(0.1)



import typing
import asyncio

from . import Middleware


class ErrorMiddleware(Middleware):
    def __init__(self, error: Exception, callback: typing.Callable, *args, **kwargs):
        """
        :param error: Exception to catch
        :param callback: callback, that will handle an error if raises
                         callback must accept error, subject, message, msg parameters
                         in order to provide full data about the error case
        """
        super().__init__(*args, **kwargs)
        self.error = error
        self.callback = callback

    async def send_any(self, subject: str, message, send_func, *args, **kwargs):
        try:
            response = await send_func(subject, message, *args, **kwargs)
            return response
        except self.error as e:
            if asyncio.iscoroutinefunction(self.callback):
                await self.callback(e, subject=subject, message=message)
            else:
                self.callback(e, subject=subject, message=message)

    async def listen_any(self, msg, callback):
        try:
            response = await callback(msg)
            return response
        except self.error as e:
            if asyncio.iscoroutinefunction(self.callback):
                await self.callback(e, msg=msg)
            else:
                self.callback(e, msg=msg)



import os
import shutil
import threading
import json
import random
import time
import typing
from urllib.parse import urljoin

import requests
import websocket
from pynats import NATSClient, NATSMessage

from .exceptions import TestClientError
from .utils.helper import start_process, start_thread
from panini.managers.nats_client import Msg


# Annotations for `Session.request()`
Cookies = typing.Union[
    typing.MutableMapping[str, str], requests.cookies.RequestsCookieJar
]
Params = typing.Union[bytes, typing.MutableMapping[str, str]]
DataType = typing.Union[bytes, typing.MutableMapping[str, str], typing.IO]
TimeOut = typing.Union[float, typing.Tuple[float, float]]
FileType = typing.MutableMapping[str, typing.IO]
AuthType = typing.Union[
    typing.Tuple[str, str],
    requests.auth.AuthBase,
    typing.Callable[[requests.Request], requests.Request],
]


class HTTPSessionTestClient(requests.Session):
    __test__ = False

    def __init__(self, base_url: str = "http://127.0.0.1:8080"):
        super(HTTPSessionTestClient, self).__init__()
        self.base_url = base_url

    def request(  # type: ignore
        self,
        method: str,
        url: str,
        params: Params = None,
        data: DataType = None,
        headers: typing.MutableMapping[str, str] = None,
        cookies: Cookies = None,
        files: FileType = None,
        auth: AuthType = None,
        timeout: TimeOut = None,
        allow_redirects: bool = None,
        proxies: typing.MutableMapping[str, str] = None,
        hooks: typing.Any = None,
        stream: bool = None,
        verify: typing.Union[bool, str] = None,
        cert: typing.Union[str, typing.Tuple[str, str]] = None,
        json: typing.Any = None,
    ) -> requests.Response:
        url = urljoin(self.base_url, url)
        return super().request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )


def get_logger_files_path(folder: str = "test_logs", remove_if_exist: bool = False):
    testing_directory_path = os.getcwd()
    testing_logs_directory_path = os.path.join(testing_directory_path, folder)
    if remove_if_exist:
        if os.path.exists(testing_logs_directory_path):
            shutil.rmtree(testing_logs_directory_path)

    return testing_logs_directory_path


class TestClient:
    __test__ = False

    def __init__(
        self,
        run_panini: typing.Callable = None,
        run_panini_args: list = None,
        run_panini_kwargs: dict = None,
        panini_service_name: str = "*",
        panini_client_id: str = "*",
        logger_files_path: str = "test_logs",
        use_web_server: bool = False,
        use_web_socket: bool = False,
        base_web_server_url: str = "http://127.0.0.1:8080",
        base_nats_url: str = "nats://127.0.0.1:4222",
        socket_timeout: int = 5,
        listener_socket_timeout: int = 1,
        auto_reconnect: bool = True,
        name: str = "__".join(
            [
                "test_client",
                str(random.randint(1, 10000000)),
                str(random.randint(1, 10000000)),
            ]
        ),
        _subscribed_subjects: [str] = None,
    ):
        self.run_panini = run_panini
        self.run_panini_args = run_panini_args or []
        self.run_panini_kwargs = run_panini_kwargs or {}
        self._subscribed_subjects = _subscribed_subjects or []
        self.panini_service_name = panini_service_name
        self.panini_client_id = panini_client_id
        self.logger_files_path = logger_files_path
        self.base_web_server_url = base_web_server_url
        self.use_web_server = use_web_server
        self.use_web_socket = use_web_socket
        self.name = name
        self.base_nats_url = base_nats_url
        self.auto_reconnect = auto_reconnect
        self.nats_client_sender = self.create_nats_client("sender", socket_timeout)
        self.nats_client_listener = self.create_nats_client(
            "listener", listener_socket_timeout
        )
        self.nats_client_sender.connect()
        self.nats_client_listener.connect()
        self.nats_client_listener_thread = None
        self.nats_client_listener_thread_stop_event = threading.Event()

        if use_web_server:
            self.http_session = HTTPSessionTestClient(base_url=base_web_server_url)

        if use_web_socket:
            self.websocket_session = websocket.WebSocket()

        self.panini_process = None

    def create_nats_client(self, suffix: str, nats_timeout) -> NATSClient:
        return NATSClient(
            url=self.base_nats_url,
            name=self.name + suffix,
            socket_timeout=nats_timeout,
        )

    @staticmethod
    def _dict_to_bytes(message: dict) -> bytes:
        return json.dumps(message).encode("utf-8")

    @staticmethod
    def _bytes_to_dict(payload: bytes) -> dict:
        return json.loads(payload)

    @staticmethod
    def wrap_run_panini(
        run_panini,
        run_panini_args: list,
        run_panini_kwargs: dict,
        logger_files_path: str,
        use_error_middleware: bool,
    ):
        from .utils.logger import get_logger

        test_logger = get_logger("panini")
        # set the panini testing data in os.environ
        os.environ["PANINI_TEST_MODE"] = "true"
        os.environ["PANINI_TEST_MODE_USE_ERROR_MIDDLEWARE"] = (
            "true" if use_error_middleware else "false"
        )
        testing_logger_files_path = (
            get_logger_files_path(logger_files_path)
            if not os.path.isabs(logger_files_path)
            else logger_files_path
        )

        os.environ["PANINI_TEST_LOGGER_FILES_PATH"] = testing_logger_files_path
        try:
            run_panini(*run_panini_args, **run_panini_kwargs)
        except Exception as e:
            test_logger.exception(f"Run panini error: {e}")

    def start(
        self,
        is_daemon: bool = True,
        do_always_listen: bool = True,
        use_error_middleware=True,
    ):
        if do_always_listen and len(self._subscribed_subjects) > 0:

            def nats_listener_worker(stop_event):
                try:
                    while not stop_event.wait(1):
                        self.nats_client_listener.wait()
                except Exception:
                    pass

            # Run nats-listener process
            self.nats_client_listener_thread = start_thread(
                nats_listener_worker,
                args=[self.nats_client_listener_thread_stop_event],
                daemon=True,
            )

        if self.run_panini is not None:

            def panini_started(msg):
                pass

            sub = self.nats_client_sender.subscribe(
                f"panini_events.{self.panini_service_name}.{self.panini_client_id}.started",
                callback=panini_started,
                max_messages=1,
            )
            self.nats_client_sender.auto_unsubscribe(sub)

            self.panini_process = start_process(
                self.wrap_run_panini,
                args=(
                    self.run_panini,
                    self.run_panini_args,
                    self.run_panini_kwargs,
                    self.logger_files_path,
                    use_error_middleware,
                ),
                daemon=is_daemon,
            )

            try:
                self.nats_client_sender.wait(count=1)  # wait for panini to start
            except OSError:
                raise TestClientError(
                    "TestClient was waiting panini to start, but panini does not started"
                )

            if self.use_web_server:
                pass  # TODO: understand, why don't we need to wait for web_server

        return self

    def stop(self):
        self.nats_client_listener.close()
        self.nats_client_sender.close()
        self.nats_client_listener_thread_stop_event.set()

        if self.panini_process:
            self.panini_process.kill()

        if hasattr(self, "http_session"):
            self.http_session.close()

        if hasattr(self, "websocket_session"):
            self.websocket_session.close()

    def publish(self, subject: str, message: dict, reply: str = "") -> None:
        self.nats_client_sender.publish(
            subject=subject, payload=self._dict_to_bytes(message), reply=reply
        )

    def reconnect(self):
        if (
            self.auto_reconnect
        ):  # reconnects nats_client for next call, if it was disconnected earlier
            try:
                self.nats_client_sender.ping()
            except Exception:
                self.nats_client_sender.reconnect()

    def request(self, subject: str, message: dict) -> dict:
        self.reconnect()

        return self._bytes_to_dict(
            self.nats_client_sender.request(
                subject=subject, payload=self._dict_to_bytes(message)
            ).payload
        )

    def subscribe(
        self,
        subject: str,
        callback: typing.Callable,
        queue: str = "",
        max_messages: typing.Optional[int] = None,
    ):
        self._subscribed_subjects.append(subject)
        return self.nats_client_listener.subscribe(
            subject=subject,
            callback=callback,
            queue=queue,
            max_messages=max_messages,
        )

    def wait(self, count: int) -> None:
        if self.nats_client_listener_thread is not None:
            raise TestClientError(
                "You can't use client.wait, "
                "if you don't have any @client.listen functions "
                "or you don't use do_always_listen=False!"
            )
        self.nats_client_listener.wait(count=count)

    def listen(self, subject: str):
        def decorator(func):
            assert isinstance(subject, str), "Subject must be only in str format"

            def wrapper(incoming_message):
                assert isinstance(incoming_message, NATSMessage)
                incoming_message_data = self._bytes_to_dict(incoming_message.payload)

                msg = Msg(
                    subject=incoming_message.subject,
                    data=incoming_message_data,
                    reply=incoming_message.reply,
                    sid=incoming_message.sid,
                )
                wrapper_response = func(msg)
                if wrapper_response is not None and incoming_message.reply != "":
                    self.nats_client_listener.publish(
                        subject=incoming_message.reply,
                        payload=self._dict_to_bytes(wrapper_response),
                    )

            self.subscribe(subject, wrapper)

            return wrapper

        return decorator



import asyncio
import json
import os
import uuid
import subprocess
import sys
import threading
import random

from multiprocessing import Process
from threading import Thread


def get_app_root_path():
    root_path = os.path.dirname(sys.argv[0])
    return f"{root_path}/" if root_path else ""


def create_dir_when_none(dir_name):
    """Check if a directory exist or create one.
    return: bool."""
    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)
        return False
    else:
        return True


async def run_coro_threadsafe(coro, other_loop, our_loop=None, many=False):
    """Schedules coro in other_loop, awaits until coro has run and returns
    its result.
    """
    loop = our_loop or asyncio.get_event_loop()
    fut = asyncio.run_coroutine_threadsafe(coro, other_loop)
    # fut = other_loop.call_soon_threadsafe(coro)
    # set up a threading.Event that fires when the future is finished
    finished = threading.Event()

    def fut_finished_cb(_):
        finished.set()

    fut.add_done_callback(fut_finished_cb)
    # wait on that event in an executor, yielding control to our_loop
    await loop.run_in_executor(None, finished.wait)
    # coro's result is now available in the future object
    return fut.result()


def start_thread(method, args=None, daemon=False):
    kwargs = dict(target=method)
    if args is not None:
        kwargs["args"] = args
    thread = Thread(**kwargs)
    if daemon:
        thread.setDaemon(True)
    thread.start()
    return thread


def start_process(method, args=None, kwargs=None, daemon=True):
    proc_kwargs = dict(target=method, daemon=daemon)
    if args is not None:
        proc_kwargs["args"] = args
    if kwargs is not None:
        proc_kwargs["kwargs"] = kwargs
    proc = Process(**proc_kwargs)
    proc.start()
    return proc


def is_json(myjson):
    try:
        json.loads(myjson)
    except Exception:
        return False
    return True


def _exec(*command, stdout_on=False, cwd=None):
    if stdout_on:
        result = subprocess.check_output(command).decode("utf-8")
        return result
    else:
        if cwd:
            subprocess.Popen(command, cwd=cwd)
        else:
            subprocess.Popen(command)


def create_client_code_by_hostname(name: str):
    return "__".join(
        [
            name,
            os.environ["HOSTNAME"]
            if "HOSTNAME" in os.environ
            else "non_docker_env_" + str(random.randint(1, 1000000)),
            str(random.randint(1, 1000000)),
        ]
    )


def validate_msg(message):
    if type(message) is dict:
        return True
    elif type(message) is str and is_json(message):
        return True
    return False


def add_isr_id_if_absent(message, isr_id: str = None):
    if "isr-id" not in message:
        if isr_id is None:
            message["isr-id"] = str(uuid.uuid4())
        else:
            message["isr-id"] = isr_id
    return message


def register_msg(message, isr_id: str = None):
    if type(message) is str and is_json(message):
        message = json.loads(message)
    return json.dumps(add_isr_id_if_absent(message, isr_id))



from .utils.logger import Logger
from nats.aio.errors import ErrTimeout

app_logger = None


class BaseError(Exception):
    pass


class BaseLoggedError(Exception):
    def __init__(self, error_message="", log_obj=None, exception_obj=None):
        if log_obj is not None:
            if not isinstance(log_obj, Logger):
                print(
                    f"Wrong log_obj for error {error_message}. It has to be Logger instance"
                )
            else:
                log_obj.error(error_message)
        else:
            global app_logger
            if app_logger is None:
                from .app import _app

                app_logger = _app.logger
                app_logger.error(error_message)
            else:
                if not isinstance(app_logger, Logger):
                    print(
                        f"Wrong log_obj for error {error_message}. It has to be Logger instance"
                    )
                else:
                    app_logger.error(error_message)
        if exception_obj:
            raise exception_obj(error_message)
        else:
            raise Exception(error_message)


class InitializingNATSError(BaseError):
    pass


class InitializingLoggerError(BaseError):
    pass


class InitializingEventManagerError(BaseError):
    pass


class InitializingValidatorError(BaseError):
    pass


class InitializingTaskError(BaseError):
    pass


class InitializingIntervalTaskError(BaseError):
    pass


class InitializingMainJobError(BaseError):
    pass


class NotReadyError(BaseError):
    pass


class NATSTimeoutError(ErrTimeout):
    pass


class EventHandlingError(BaseError):
    pass


class PublishError(BaseError):
    pass


class RequestError(BaseError):
    pass


class ValidationError(Exception):
    pass


class DecodeError(BaseError):
    pass


class DataTypeError(BaseError):
    pass


class TestClientError(BaseError):
    pass



import json
import os
import threading
import time
import collections
from datetime import datetime
from typing import Any

from panini import middleware
from panini.app import get_app


class _Writer(threading.Thread):
    def __init__(self, filename: str):
        threading.Thread.__init__(self)

        self._array = collections.deque()
        self._lock = threading.Lock()

        self._filename = filename

    def add(self, event: dict):
        with self._lock:
            self._array.append(event)

    def run(self):

        """
        This is a main method of _Writer.
        It checks the array in order to write it to the file with events

        It keeps the file opened if there are any messages, to do not spend time for reopen every time

        Also, it has gradually increasing sleep timer

        """

        file = None
        min_sleep_time = 0.1
        max_sleep_time = 1
        sleep_time = min_sleep_time
        while True:
            try:
                if len(self._array) == 0:
                    # close file only when array is empty
                    if file:
                        file.close()
                        file = None

                    # gradually increase sleep time, in order to handle appeared massages faster
                    time.sleep(sleep_time)
                    sleep_time = min(max_sleep_time, sleep_time * 2)
                else:
                    if not file:
                        file = open(self._filename, "a")

                    line = self._array.popleft()
                    file.write(json.dumps(line) + os.linesep)
                    sleep_time = min_sleep_time
            except Exception as ex:
                print(f"exception: {ex}")


class WriterEmulatorMiddleware(middleware.Middleware):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)

        app = get_app()
        print("Emulator is experimental feature!")
        folder = kwargs.get("folder")
        if not os.path.isdir(folder):
            os.makedirs(folder)

        filename = f"{folder}/events.{app.service_name}.{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}.jsonl"
        self._prefix = kwargs.get("prefix", "emulator")
        self._writer = _Writer(filename)
        self._writer.start()

    async def send_publish(self, subject: str, message: Any, publish_func, **kwargs):

        self._writer.add(
            {
                "event_type": "send_publish",
                "subject": subject,
                "message": message,
                "timestamp": time.time(),
            }
        )

        await publish_func(subject, message, **kwargs)

    async def send_request(self, subject: str, message: Any, request_func, **kwargs):
        response = await request_func(subject, message, **kwargs)

        self._writer.add(
            {
                "event_type": "send_request",
                "subject": subject,
                "message": message,
                "response": response,
                "timestamp": time.time(),
            }
        )

        return response

    async def listen_publish(self, msg, callback):

        self._writer.add(
            {
                "event_type": "listen_publish",
                "subject": msg.subject,
                "message": msg.data,
                "context": msg.context,
                "timestamp": time.time(),
            }
        )

        await callback(msg)

    async def listen_request(self, msg, callback):

        response = await callback(msg)

        self._writer.add(
            {
                "event_type": "listen_request",
                "subject": msg.subject,
                "message": msg.data,
                "context": msg.context,
                "response": response,
                "timestamp": time.time(),
            }
        )

        return response



from aiohttp import web
from aiohttp.web_routedef import RouteTableDef


class HTTPServer:
    def __init__(
        self,
        routes: RouteTableDef,
        host: str = None,
        port: int = None,
        web_app: web.Application = None,
        web_server_params=None,
    ):
        if web_server_params is None:
            web_server_params = {}

        self.routes = routes
        self.host = host
        self.port = port
        self.web_server_params = web_server_params

        if web_app:
            self.web_app = web_app
        else:
            self.web_app = web.Application()

    def start_server(self):
        self._start_server()

    def _start_server(self):
        self.web_app.add_routes(self.routes)
        web.run_app(self.web_app, host=self.host, port=self.port, **self.web_server_params)



import uuid

from panini.app import get_app
import copy

from panini.emulator_client import EmulatorClient
from panini.middleware import Middleware
from panini.managers.nats_client import Msg
from panini.utils.helper import start_thread


class ReaderEmulatorMiddleware(Middleware):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)

        self._prefix = kwargs.get("prefix", str(uuid.uuid4()))

        app = get_app()
        assert app is not None
        print("Emulator is experimental feature!")
        subjects = copy.copy(list(app.SUBSCRIPTIONS.keys()))
        for subject in subjects:
            subject_with_prefix = self._prefix + "." + subject
            app.SUBSCRIPTIONS[subject_with_prefix] = app.SUBSCRIPTIONS.pop(subject)

        self._run_emulator = kwargs.get("run_emulator", True)

        if self._run_emulator:
            emulator = EmulatorClient(
                prefix=self._prefix,
                filepath=kwargs.get("filename"),
                emulate_timeout=kwargs.get("emulate_timeout", True),
                compare_output=kwargs.get("compare_output", False),
                app_name=app.service_name,
            )
            start_thread(emulator.start)
            emulator.wait_for_readiness()

    async def listen_any(self, message: Msg, callback):
        message.subject = message.subject[len(self._prefix) + 1 :]
        response = await callback(message)
        return response

    async def send_any(self, subject: str, message, send_func, *args, **kwargs):
        subject = self._prefix + "." + subject
        response = await send_func(subject, message, *args, **kwargs)
        return response



import typing
import ujson
import asyncio
import threading
import nest_asyncio
from types import CoroutineType
from typing import Union, List
from nats.aio.client import Client as NATS
from panini.exceptions import DataTypeError
from panini.managers.middleware_manager import MiddlewareManager
from panini.utils.logger import get_logger

NoneType = type(None)
nest_asyncio.apply()


class Msg:
    """
    Alternative implementation of the class with "context" field
    """

    __slots__ = ("subject", "reply", "data", "sid", "context")

    def __init__(self, subject="", reply="", data=b"", sid=0, context=None):
        if context is None:
            context = {}
        self.subject = subject
        self.reply = reply
        self.data = data
        self.sid = sid
        self.context = context

    def __repr__(self):
        return "<{}: subject='{}' reply='{}' context='{}...'>".format(
            self.__class__.__name__,
            self.subject,
            self.reply,
            self.context,
        )


class NATSClient:
    def __init__(
            self,
            host: Union[str, NoneType],
            port: Union[int, str],
            servers: Union[List[str], NoneType],
            client_id: str,
            allow_reconnect: Union[bool, NoneType],
            max_reconnect_attempts: int = 60,
            reconnecting_time_wait: int = 2,
            auth: Union[dict, NoneType] = None,
            queue="",
            pending_bytes_limit=65536 * 1024 * 10,
            **kwargs
    ):
        """
        :param client_id: instance identifier for NATS, str
        :param port: default '4222'
        :param allow_reconnect: False if you want to stop instance when connection lost
        :param max_reconnect_attempts:
        :param reconnecting_time_wait:
        """
        if auth is None:
            auth = {}
        self.log = get_logger("panini")
        self.connected = False
        self.client_id = client_id
        self.host = host
        self.port = port
        self.servers = servers
        self.queue = queue
        self.auth = auth
        self.listen_subjects_callbacks = None
        self.allow_reconnect = allow_reconnect
        self.max_reconnect_attempts = max_reconnect_attempts
        self.reconnecting_time_wait = reconnecting_time_wait
        self.pending_bytes_limit = pending_bytes_limit
        self.ssid_map = {}

        self.include_subjects = None
        self.exclude_subjects = None

        self._middleware_manager = MiddlewareManager()

        def not_assigned_method(*args, **kwargs):
            raise Exception("used before assignment")

        self._publish_wrapped = not_assigned_method
        self._request_wrapped = not_assigned_method

        self._connection_kwargs = kwargs

        self.loop = asyncio.get_event_loop()

    def set_listen_subjects_callbacks(self, subscriptions: list):
        subscriptions = self.filter_subjects(subscriptions)
        self.listen_subjects_callbacks = subscriptions

    def start(self):
        # inject send_middlewares
        self._publish_wrapped = self._middleware_manager.wrap_function_by_middleware(
            "publish"
        )(self._publish)
        self._request_wrapped = self._middleware_manager.wrap_function_by_middleware(
            "request"
        )(self._request)

        self.loop.run_until_complete(self._establish_connection())

    @property
    def middleware_manager(self):
        return self._middleware_manager

    @property
    def middlewares(self):
        return self._middleware_manager.middlewares

    @middlewares.setter
    def middlewares(self, value: dict):
        self._middleware_manager.middlewares = value

    async def _establish_connection(self):
        self.client = NATS()
        self.client.msg_class = Msg

        if self.servers is None:
            server = 'nats://' + self.host + ":" + str(self.port)
            self.servers = [server]

        kwargs = {"servers": self.servers, "loop": self.loop, "name": self.client_id, **self._connection_kwargs}
        if self.allow_reconnect:
            kwargs["allow_reconnect"] = self.allow_reconnect
        if self.max_reconnect_attempts:
            kwargs["max_reconnect_attempts"] = self.max_reconnect_attempts
        if self.reconnecting_time_wait:
            kwargs["reconnect_time_wait"] = self.reconnecting_time_wait
        kwargs.update(self.auth)
        await self.client.connect(**kwargs)
        if self.client.is_connected:
            listen_subjects_callbacks = self.listen_subjects_callbacks
            for subject, callbacks in listen_subjects_callbacks.items():
                for callback in callbacks:
                    await self.subscribe_new_subject(
                        subject, callback, init_subscription=True
                    )

    def subscribe_new_subject_sync(self, subject: str, callback: CoroutineType, **kwargs):
        self.loop.run_until_complete(self.subscribe_new_subject(subject, callback, **kwargs))

    async def subscribe_new_subject(
            self,
            subject: str,
            callback: CoroutineType,
            init_subscription=False,
            is_async=False,
            data_type=None
    ):
        if data_type == None:
            data_type = getattr(callback, 'data_type', 'json.loads')

        callback = self._middleware_manager.wrap_function_by_middleware("listen")(callback)
        wrapped_callback = _ReceivedMessageHandler(self._publish, callback, data_type)
        ssid = await self.client.subscribe(
            subject,
            queue=self.queue,
            cb=wrapped_callback,
            pending_bytes_limit=self.pending_bytes_limit,
            is_async=is_async,
        )
        if subject not in self.ssid_map:
            self.ssid_map[subject] = []
        self.ssid_map[subject].append(ssid)
        if init_subscription is False:
            if subject not in self.listen_subjects_callbacks:
                self.listen_subjects_callbacks[subject] = []
            self.listen_subjects_callbacks[subject].append(callback)
        return ssid

    def unsubscribe_subject_sync(self, subject: str):
        self.loop.run_until_complete(self.unsubscribe_subject(subject))

    async def unsubscribe_subject(self, subject: str):
        if subject not in self.ssid_map:
            raise Exception(f"Subject {subject} hasn't been subscribed")
        for ssid in self.ssid_map[subject]:
            await self.client.unsubscribe(ssid)
        del self.ssid_map[subject]
        del self.listen_subjects_callbacks[subject]

    def unsubscribe_ssid_sync(self, ssid: int, subject: str = None):
        self.loop.run_until_complete(self.unsubscribe_ssid(ssid, subject))

    async def unsubscribe_ssid(self, ssid: int, subject: str = None):
        if subject and subject not in self.ssid_map:
            raise Exception(f"Subject {subject} hasn't been subscribed")
        await self.client.unsubscribe(ssid)
        if subject:
            del self.ssid_map[subject]
            del self.listen_subjects_callbacks[subject]
        else:
            for subject in self.ssid_map:
                if ssid in self.ssid_map[subject]:
                    self.ssid_map[subject].remove(ssid)
            for subject in self.listen_subjects_callbacks:
                if ssid in self.listen_subjects_callbacks[subject]:
                    self.listen_subjects_callbacks[subject].remove(ssid)

    def publish_sync(
            self,
            subject: str,
            message,
            reply_to: str = None,
            force: bool = False,
            data_type: type or str = "json.dumps",
    ):
        asyncio.ensure_future(
            self.publish(subject, message, reply_to, force, data_type)
        )

    def publish_from_another_thread(self, subject: str, message):
        self.loop.call_soon_threadsafe(self.publish_sync, subject, message)

    def request_sync(
            self,
            subject: str,
            message,
            timeout: int = 10,
            data_type: type or str = "json.dumps",
            callback: typing.Callable = None,
    ):
        # asyncio.ensure_future(self.request(subject, message, timeout, data_type))
        return self.loop.run_until_complete(
            self.request(subject, message, timeout, data_type, callback)
        )

    def request_from_another_thread_sync(
            self,
            subject: str,
            message,
            timeout: int = 10,
    ):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
        return loop.run_until_complete(
            self.request_from_another_thread(subject, message, timeout)
        )

    async def request_from_another_thread(
            self,
            subject: str,
            message,
            timeout: int = 10,
    ):
        # return self.loop.call_soon_threadsafe(self.request_sync, subject, message, timeout)
        fut = asyncio.run_coroutine_threadsafe(
            self.request(subject, message, timeout), self.loop
        )
        finished = threading.Event()

        def fut_finished_cb(_):
            finished.set()

        fut.add_done_callback(fut_finished_cb)
        await asyncio.get_event_loop().run_in_executor(None, finished.wait)
        return fut.result()

    @staticmethod
    def format_message_data_type(message, data_type):
        if type(message) in [dict, list] and data_type == "json.dumps":
            message = ujson.dumps(message)
            message = message.encode()
        elif type(message) is str and data_type is str:
            message = message.encode()
        elif type(message) is bytes and data_type is bytes:
            pass
        else:
            raise DataTypeError(
                f'Expected {"dict" if data_type in [dict, "json.dumps"] else data_type} but got {type(message)}'
            )

        return message

    async def _publish(
            self,
            subject: str,
            message,
            reply_to: str = None,
            force: bool = False,
            data_type: type or str = "json.dumps",
    ):
        message = self.format_message_data_type(message, data_type)

        if reply_to is not None:
            await self.client.publish_request(subject, reply_to, message)
        else:
            await self.client.publish(subject, message)
        if force:
            await self.client.flush()

    async def publish(
            self,
            subject: str,
            message,
            reply_to: str = None,
            force: bool = False,
            data_type: type or str = "json.dumps",
    ):
        return await self._publish_wrapped(
            subject=subject,
            message=message,
            reply_to=reply_to,
            force=force,
            data_type=data_type,
        )

    async def _request(
            self,
            subject: str,
            message,
            timeout: int = 10,
            data_type: type or str = "json.dumps",
            callback: typing.Callable = None,
    ):
        message = self.format_message_data_type(message, data_type)
        if callback is not None:
            return await self.client.request(subject, message, cb=callback)
        response = await self.client.request(subject, message, timeout=timeout)
        response = response.data
        if data_type == "json.dumps":
            response = ujson.loads(response)
        elif data_type is str:
            response = response.decode()
        return response

    async def request(
            self,
            subject: str,
            message,
            timeout: int = 10,
            data_type: type or str = "json.dumps",
            callback: typing.Callable = None,
    ):
        return await self._request_wrapped(
            subject=subject,
            message=message,
            timeout=timeout,
            data_type=data_type,
            callback=callback,
        )

    def disconnect_sync(self):
        self.loop.run_until_complete(self.disconnect())

    async def disconnect(self):
        await self.client.drain()
        self.log.warning("Disconnected")

    def check_connection(self):
        if self.client._status is NATS.CONNECTED:
            self.log.info("NATS Client status: CONNECTED")
            return True
        self.log.warning("NATS Client status: DISCONNECTED")

    def add_filters(self, include: list = None, exclude: list = None):
        self.include_subjects = include
        self.exclude_subjects = exclude

    def filter_subjects(self, subscriptions):
        assert not self.include_subjects or not self.exclude_subjects, "You can use either include or exclude. Not both"

        if self.include_subjects:
            for subject in subscriptions.copy():
                success = False
                for subject_include in self.include_subjects:
                    if subject_include in subject:
                        success = True
                        break
                if not success:
                    del subscriptions[subject]

        if self.exclude_subjects:
            for subject in subscriptions.copy():
                for subject_exclude in self.exclude_subjects:
                    if subject_exclude in subject:
                        del subscriptions[subject]
                        break

        return subscriptions


class _ReceivedMessageHandler:
    def __init__(self, publish_func, cb, data_type):
        self.publish_func = publish_func
        self.cb = cb
        self.data_type = data_type
        self.cb_is_async = asyncio.iscoroutinefunction(cb)

    def __call__(self, msg):
        asyncio.ensure_future(self.call(msg))

    async def call(self, msg):
        self.parse_data(msg)
        reply_to = self.match_msg_case(msg)
        if self.cb_is_async:
            response = await self.cb(msg)
        else:
            response = self.cb(msg)
        if reply_to is not None:
            await self.publish_func(reply_to, response)

    def parse_data(self, msg):
        if self.data_type == "raw" or self.data_type == bytes:
            return
        if self.data_type == str:
            msg.data = msg.data.decode()
        elif self.data_type == dict or self.data_type == "json.loads":
            msg.data = ujson.loads(msg.data.decode())
        else:
            raise Exception(f"{self.data_type} is unsupported data format")

    def match_msg_case(self, msg):
        if not msg.reply == "":
            reply_to = msg.reply
        elif self.data_type == "json" and "reply_to" in msg.data:
            reply_to = msg.data.pop("reply_to")
        else:
            reply_to = None
        return reply_to



import asyncio
from typing import Callable, Type

from panini.middleware import Middleware


class MiddlewareManager:

    def __init__(self):
        self._middlewares = {
            "send_publish_middleware": [],
            "listen_publish_middleware": [],
            "send_request_middleware": [],
            "listen_request_middleware": [],
        }

    @property
    def middlewares(self):
        return self._middlewares

    @middlewares.setter
    def middlewares(self, value: dict):
        self._middlewares = value

    def add_middleware(self, middleware_cls: Type[Middleware], *args, **kwargs):
        assert issubclass(middleware_cls, Middleware), \
            "Each custom middleware class must be a subclass of Middleware"

        high_priority_functions = (
            "send_publish",
            "listen_publish",
            "send_request",
            "listen_request",
        )

        global_functions = (
            "send_any",
            "listen_any"
        )

        # check, that at least one function is implemented
        assert any(
            function in middleware_cls.__dict__
            for function in high_priority_functions + global_functions
        ), f"At least one of the following functions must be implemented: {high_priority_functions + global_functions}"

        middleware_obj = middleware_cls(*args, **kwargs)
        for function_name in high_priority_functions:
            if function_name in middleware_cls.__dict__:
                self._middlewares[f"{function_name}_middleware"].append(
                    getattr(middleware_obj, function_name)
                )

        if "send_any" in middleware_cls.__dict__:
            if "send_publish" not in middleware_cls.__dict__:
                self._middlewares["send_publish_middleware"].append(
                    middleware_obj.send_any
                )

            if "send_request" not in middleware_cls.__dict__:
                self._middlewares["send_request_middleware"].append(
                    middleware_obj.send_any
                )

        if "listen_any" in middleware_cls.__dict__:
            if "listen_publish" not in middleware_cls.__dict__:
                self._middlewares["listen_publish_middleware"].append(
                    middleware_obj.listen_any
                )

            if "listen_request" not in middleware_cls.__dict__:
                self._middlewares["listen_request_middleware"].append(
                    middleware_obj.listen_any
                )

    def wrap_function_by_middleware(self, function_type: str) -> Callable:
        """
        function: Callable - function to wrap
        function_type: str - one of the ["listen", "publish", "request"]
        """
        assert function_type in (
            "listen",
            "publish",
            "request",
        ), "function type must be in (`listen`, `publish`, `request`)"

        def decorator(function: Callable):
            def wrap_function_by_send_middleware(
                func: Callable, single_middleware
            ) -> Callable:
                def next_wrapper(subject: str, message, *args, **kwargs):
                    return single_middleware(subject, message, func, *args, **kwargs)

                async def async_next_wrapper(subject: str, message, *args, **kwargs):
                    return await single_middleware(
                        subject, message, func, *args, **kwargs
                    )

                if asyncio.iscoroutinefunction(single_middleware):
                    return async_next_wrapper
                else:
                    return next_wrapper

            def wrap_function_by_listen_middleware(
                func: Callable, single_middleware
            ) -> Callable:
                def next_wrapper(msg):
                    return single_middleware(msg, func)

                async def async_next_wrapper(msg):
                    return await single_middleware(msg, func)

                if asyncio.iscoroutinefunction(single_middleware):
                    return async_next_wrapper
                else:
                    return next_wrapper

            def build_middleware_wrapper(
                func: Callable, middleware_key: str, wrapper: Callable
            ) -> Callable:
                for middleware in self._middlewares[middleware_key]:
                    func = wrapper(func, middleware)
                return func

            if function_type == "publish":
                return build_middleware_wrapper(
                    function,
                    "send_publish_middleware",
                    wrap_function_by_send_middleware,
                )

            elif function_type == "request":
                return build_middleware_wrapper(
                    function,
                    "send_request_middleware",
                    wrap_function_by_send_middleware,
                )

            else:
                if (
                    len(self._middlewares["listen_publish_middleware"]) == 0
                    and len(self._middlewares["listen_request_middleware"])
                    == 0
                ):
                    return function

                function_listen_publish = build_middleware_wrapper(
                    function,
                    "listen_publish_middleware",
                    wrap_function_by_listen_middleware,
                )
                function_listen_request = build_middleware_wrapper(
                    function,
                    "listen_request_middleware",
                    wrap_function_by_listen_middleware,
                )

                def listen_wrapper(msg) -> Callable:
                    if msg.reply == "":
                        return function_listen_publish(msg)
                    else:
                        return function_listen_request(msg)

                async def async_listen_wrapper(msg) -> Callable:
                    if msg.reply == "":
                        return await function_listen_publish(msg)
                    else:
                        return await function_listen_request(msg)

                if asyncio.iscoroutinefunction(function):
                    return async_listen_wrapper
                else:
                    return listen_wrapper

        return decorator



class Middleware:
    def __init__(self, *args, **kwargs):
        pass

    async def send_publish(self, subject: str, message, publish_func, *args, **kwargs):
        """
        :param subject: str
        :param message: any of supported types
        :param publish_func: Callable for publish
        :return: None
        """
        pass

    async def listen_publish(self, msg, callback):
        """
        :param msg: Msg
        :param callback: Callable, that will be called on receive message
        :return: None
        """
        pass

    async def send_request(self, subject: str, message, request_func, *args, **kwargs):
        """
        :param subject: str
        :param message: any of supported types
        :param request_func: Callable for request
        :return: any of supported types
        """
        pass

    async def listen_request(self, msg, callback):
        """
        :param msg: Msg
        :param callback: Callable, that will be called on receive message
        :return: any of supported types
        """
        pass

    async def send_any(self, subject: str, message, send_func, *args, **kwargs):
        """
        :param subject: str
        :param message: any of supported types
        :param send_func: Callable for send
        :return: None or any of supported types
        """

    async def listen_any(self, msg, callback):
        """
        :param msg: Msg
        :param callback: Callable, that will be called on receive message
        :return: None or any of supported types
        """



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

long_description = """
Panini is a modern framework for quick development of streaming microservices.
Our goal is to create fastapi/aiohttp/flask-like solution but for NATS streaming.
The framework allows you to work with NATS features and some additional logic using a simple interface:

- publish messages to subject
- subscribe to subject
- request-response
- request-response to another subject
- tasks
- periodic tasks
- middlewares
- HTTP server

"""

setup(
    name="panini",
    version="0.6.0b1",
    description="A python messaging framework for microservices based on NATS",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/lwinterface/panini",
    author="Op Return SA, developers: Andrii Volotskov, Danylo Tiutiushkin, Oleksii Volotskov",
    author_email="example@example.com",
    python_requires=">=3.8.3",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: System :: Networking",
        "Topic :: System :: Distributed Computing",
    ],
    packages=[
        "panini",
        "panini.utils",
        "panini.http_server",
        "panini.middleware",
        "panini.managers",
    ],
    install_requires=[
        "aiohttp>=3.7.4",
        "aiohttp-cors>=0.7.0",
        "async-timeout>=3.0.1",
        "asyncio-nats-client>=0.11.2",
        "websocket-client>=0.58.0",
        "requests>=2.24.0",
        "six>=1.15.0",
        "yarl>=1.6.1",
        "python-json-logger>=2.0.1",
        "nest-asyncio==1.5.1",
        "prometheus-client==0.9.0",
        "nats-python>=0.8.0",
        "ujson==4.0.2",
    ],
    keywords=[
        "nats",
        "microservice",
        "stream",
        "processing",
        "asyncio",
        "distributed",
        "queue",
    ],
    project_urls={  # Optional
        "Bug Reports": "https://github.com/lwinterface/panini/issues",
        "Source": "https://github.com/lwinterface/panini/",
    },
    zip_safe=False,
)



import json
import logging
import logging.handlers
import logging.config
import os
import datetime
from importlib import reload
from multiprocessing import Process, Queue, Event


from ..utils import helper


# Decorator for check, if logger exist
def check_logger_connected(func):
    def wrapper(self, msg, **extra):
        if self.logger is None:
            raise Exception("Logger hasn't been connected")
        else:
            func(self, msg, **extra)

    return wrapper


class Logger:
    """Generate logging systems which can be simply customized by adding config/log_config.json file to app_root_path
    self.logger - use built in logging system but with improved interface for logging
    """

    def __init__(self, logger: logging.Logger):
        self.logger = logger

    @check_logger_connected
    def debug(self, msg, **extra):
        self.logger.debug(msg, extra={"extra": extra})

    @check_logger_connected
    def info(self, msg, **extra):
        self.logger.info(msg, extra={"extra": extra})

    @check_logger_connected
    def warning(self, msg, **extra):
        self.logger.warning(msg, extra={"extra": extra})

    @check_logger_connected
    def error(self, msg, **extra):
        self.logger.error(msg, extra={"extra": extra})

    @check_logger_connected
    def exception(self, msg, **extra):
        self.logger.exception(msg, extra={"extra": extra})


def set_logger(
    ms_name: str,
    app_root_path: str,
    logger_files_path: str,
    in_separate_process: bool,
    client_id: str = None,
):
    logger_config = _get_logger_config(
        app_root_path, logger_files_path, ms_name, client_id
    )

    if in_separate_process:
        (
            logger_queue,
            log_stop_event,
            log_process
        ) = _set_log_recorder_process(logger_config)
        _set_main_logging_config(logger_queue)
        return (
            logger_queue,
            log_stop_event,
            log_process
        )
    else:
        logging.config.dictConfig(logger_config)
        return


def get_logger(name: str = "panini") -> Logger:
    return Logger(logging.getLogger(name))


def _get_logger_config(
    app_root_path: str, logger_files_path: str, ms_name: str, client_id: str = None
):
    if os.path.isabs(logger_files_path):
        log_dir_path = logger_files_path
    else:
        log_dir_path = os.path.join(app_root_path, logger_files_path)
    custom_config_path = os.path.join(app_root_path, "config", "log_config.json")
    if os.path.exists(custom_config_path):
        config = _configure_logging_with_custom_config_file(custom_config_path)

    else:
        config = _configure_default_logging(ms_name)
    config["handlers"] = _modify_handlers(
        config["handlers"], log_dir_path, ms_name=ms_name, client_id=client_id
    )
    return config


def _replace_keywords(filename: str, ms_name: str = None, client_id: str = None):
    """replace keywords with some meaningful data"""
    keywords = {
        "%MS_NAME%": ms_name,
        "%CLIENT_ID%": client_id,
        "%DATETIME%": str(datetime.datetime.now()),
    }
    for keyword, meaningful_data in keywords.items():
        filename = filename.replace(keyword, meaningful_data)

    return filename


def _modify_handlers(
    handlers, log_dir_path, ms_name: str = None, client_id: str = None
):
    for handler in handlers:
        if handler == "console":
            continue

        filename = _replace_keywords(handlers[handler]["filename"], ms_name, client_id)

        if not os.path.isabs(filename):
            filename = os.path.join(log_dir_path, filename)

        helper.create_dir_when_none(os.path.dirname(filename))
        handlers[handler]["filename"] = filename

    return handlers


def _basic_file_handler_skeleton(name: str, level="DEBUG", formatter="detailed"):
    return {
        "level": level,
        "class": "logging.handlers.RotatingFileHandler",
        "filename": f"{name}.log",
        "mode": "a",
        "formatter": formatter,
        "maxBytes": 1000000,
        "backupCount": 10,
    }


def _configure_default_logging(name):
    default_log_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "detailed": {
                "class": "pythonjsonlogger.jsonlogger.JsonFormatter",
                # do not remove 'extra' field - that's for extra arguments logging
                "format": "%(created)f %(name)s %(levelname)s %(processName)s %(threadName)s %(message)s %(extra)s",
            },
            "simple": {
                "class": "logging.Formatter",
                "format": "%(asctime)s %(name)-15s %(levelname)-8s %(message)s",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "simple",
                "stream": "ext://sys.stdout",
            },
            "panini": _basic_file_handler_skeleton("panini"),
            "inter_services_request": _basic_file_handler_skeleton(
                "inter_services_request"
            ),
            "app": _basic_file_handler_skeleton("app"),
            "errors": {
                "class": "logging.FileHandler",
                "filename": "errors.log",
                "mode": "a",
                "level": "ERROR",
                "formatter": "detailed",
            },
        },
        "loggers": {
            "panini": {"handlers": ["panini"]},
            "inter_services_request": {"handlers": ["inter_services_request"]},
        },
        "root": {"level": "DEBUG", "handlers": ["console", "errors", "app"]},
    }
    if name not in ("panini", "inter_services_request"):
        default_log_config["handlers"][name] = _basic_file_handler_skeleton(name)

        default_log_config["loggers"][name] = {"handlers": [name]}

    return default_log_config


def _configure_logging_with_custom_config_file(custom_config_path) -> dict:
    try:
        with open(custom_config_path, mode="r", encoding="utf-8") as f:
            config = json.load(f)

        if "detailed" not in config["formatters"]:
            config["formatters"]["detailed"] = {
                "class": "pythonjsonlogger.jsonlogger.JsonFormatter",
                "format": "%(created)s %(name)s %(levelname)s %(processName)s %(threadName)s %(message)s %(extra)s",
            }
        for basic_config in ("panini", "inter_services_request"):
            if basic_config not in config["handlers"]:
                config["handlers"][basic_config] = _basic_file_handler_skeleton(
                    basic_config
                )
                config["loggers"][basic_config] = {"handlers": [basic_config]}

        return config

    except Exception as e:
        _emergency_logging()
        logging.exception(f"Error when loading the logging configuration: {e}")
        raise SystemExit()


class LogHandler:
    @staticmethod
    def handle(record) -> None:
        if record.name == "root":
            logger = logging.getLogger()
        else:
            logger = logging.getLogger(record.name)

        if logger.isEnabledFor(record.levelno):
            logger.handle(record)


class ChangeConfigHandler:
    @staticmethod
    def handle(record) -> None:
        print(record)
        # TODO: implement changing logger configuration during runtime


def _emergency_logging() -> None:
    logging.basicConfig(
        format="%(asctime)s %(name)s %(levelname)s %(message)s",
        handlers=[
            logging.FileHandler("logging_error.log", mode="a"),
            logging.StreamHandler(),
        ],
        level=logging.DEBUG,
    )


def _dedicated_listener_process(
    log_listener_queue: Queue,
    stop_event: Event,
    config: dict
) -> None:
    logging.config.dictConfig(config)
    log_listener = logging.handlers.QueueListener(log_listener_queue, LogHandler())
    log_listener.start()
    stop_event.wait()
    log_listener.stop()


def _set_log_recorder_process(config: dict) -> (Queue, Event, Process):
    try:
        log_listener_queue = Queue()
        stop_event = Event()
        listener_process = Process(
            target=_dedicated_listener_process,
            name="listener",
            args=(
                log_listener_queue,
                stop_event,
                config
            ),
        )
        listener_process.start()
        return (
            log_listener_queue,
            stop_event,
            listener_process
        )

    except Exception as e:
        _emergency_logging()
        logging.exception(f"Error when loading the logging configuration: {e}")
        raise SystemExit()


def _set_main_logging_config(log_listener_queue: Queue):
    logging.shutdown()
    reload(logging)
    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": {
            "queue": {
                "class": "logging.handlers.QueueHandler",
                "queue": log_listener_queue,
            }
        },
        "root": {
            "level": "DEBUG",
            "handlers": ["queue"],
        },
    }
    logging.config.dictConfig(config)



import traceback

from nats.aio.errors import ErrTimeout

from ..app import get_app
from .error import ErrorMiddleware


class NATSTimeoutMiddleware(ErrorMiddleware):
    def __init__(self, subject, app=None, send_func_type: str = "request"):
        if app is None:
            app = get_app()
        assert app is not None

        async def handle_nats_timeout_callback(error: Exception, **kwargs):
            error_msg = {"error": traceback.format_exc(), "error_msg": str(error)}
            error_msg.update(kwargs)
            if send_func_type == "request":
                response = await app.request(subject, error_msg)
                app.logger.info(f"Response from nats_timeout subject: {response}")
            else:
                await app.publish(subject, error_msg)

        super().__init__(ErrTimeout, handle_nats_timeout_callback)

    async def send_any(self, subject: str, message, send_func, *args, **kwargs):
        return await super(NATSTimeoutMiddleware, self).send_any(
            subject, message, send_func, *args, **kwargs
        )

    async def listen_any(self, msg, callback):
        return await super(NATSTimeoutMiddleware, self).listen_any(msg, callback)
